//c/c++
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>


//mz general
#include "mzparameters.h"
#include "mzfunctions.h"


//mz for FF models
#include "mzfunctions_FF_models.h"


using namespace std;


void mz_FF_model(int FF_flag, double q2, 
                double& Re_GE, double& Im_GE,
                double& Re_GM, double& Im_GM,
                double& Re_F1, double& Im_F1,
		 double& Re_F2, double& Im_F2){
  /**
   * function mz_FF_model
   *
   * input arguments:
   *  FF_flag: flag to set the FF model
   *  q2:      photon 4-momentum square)
   *
   * output arguments:
   *  Re_GE, Im_GE: real and imaginary parts of electric Sachs FF
   *  Re_GM, Im_GM: real and imaginary parts of magnetic Sachs FF
   *  Re_F1, Im_F1: real and imaginary parts of Pauli-Dirac F1 FF
   *  Re_F2, Im_F2: real and imaginary parts of Pauli-Dirac F2 FF
   *
   */


  if( !( (FF_flag==0)||(FF_flag==1)||(FF_flag==2) ) ){
    printf("ERROR: mz_FF_model called with FF flag out of range: \n");
    printf("    => FF flag should be 0 (=pQCD), 1 (=vmd) or 2 (=dipole-modified)\n");
    printf("    => job killed!!! \n");
    printf("\n\n");
    exit(1);
  }

  //set model
  if(FF_flag==0) mz_FF_pQCD(q2, Re_GE, Im_GE, Re_GM, Im_GM, Re_F1, Im_F1, Re_F2, Im_F2);
  if(FF_flag==1) mz_FF_vmd_mod(q2, Re_GE, Im_GE, Re_GM, Im_GM, Re_F1, Im_F1, Re_F2, Im_F2);
  if(FF_flag==2) mz_FF_Tomasi(q2, Re_GE, Im_GE, Re_GM, Im_GM, Re_F1, Im_F1, Re_F2, Im_F2);

}



void mz_FF_pQCD(double q2, 
                double& Re_GE, double& Im_GE,
                double& Re_GM, double& Im_GM,
                double& Re_F1, double& Im_F1,
                double& Re_F2, double& Im_F2){
  /**
   * function mz_FF_pQCD 
   *
   * calculates, for a given q2, the values of both Sachs and Pauli-Dirac 
   * proton FF, according to the "pQCD inspired" model in
   * Phys.Rev.C75:045205,2007 (Adamuscin et al.) eq.28
   * Eur. Phys. J A 44, 373-384 (2010) (Sudol et al.) eq. 8 
   *
   * input arguments: 
   *    q2: photon 4-momentum square)
   *  
   * output arguments:
   *    Re_GE, Im_GE: real and imaginary parts of electric Sachs FF 
   *    Re_GM, Im_GM: real and imaginary parts of magnetic Sachs FF
   *    Re_F1, Im_F1: real and imaginary parts of Pauli-Dirac F1 FF
   *    Re_F2, Im_F2: real and imaginary parts of Pauli-Dirac F2 FF
   *
   */

  double PI=acos(-1.0);

  //double A_p=98.0;   //fit parameter A(p) eq. 28 Adamuscin et al., (GeV^4)
  double A_p=89.45;   //fit parameter A(p) eq. 8 Sudol et al., (GeV^4)
  double Lambda=0.3; //lambda QCD scale eq. 28 paper, (GeV)
  double G;


  //check argument range
  if(q2<(Lambda*Lambda)){
    printf("ERROR: function mz_FF_pQCD called with argument q2 out of range \n");
    printf("    => q2 should be larger than the QCD scale Lambda^2 = %5.3f GeV\n", (Lambda*Lambda));
    printf("\n\n");
    exit(1);
  }


  //G=(A_p)/( q2*q2*(log(q2/Lambda*Lambda)*log(q2/Lambda*Lambda) +PI*PI) );
  G=(A_p)/( q2*q2*( log(q2/(Lambda*Lambda))*log(q2/(Lambda*Lambda)) +PI*PI) );

  Re_GE=G;
  Im_GE=0.0;

  Re_GM=G;   
  Im_GM=0.0;

  Re_F1=G;
  Im_F1=0.0;

  Re_F2=0.0;
  Im_F2=0.0;


}



void mz_FF_vmd(double q2, 
               double& Re_GE, double& Im_GE,
               double& Re_GM, double& Im_GM,
               double& Re_F1, double& Im_F1,
               double& Re_F2, double& Im_F2){
  /**
   * function mz_FF_pQCD 
   *
   * calculates, for a given q2, the values of both Sachs and Pauli-Dirac 
   * proton FF in the time-like region, 
   * according to the "vector meson dominance" model in
   * Phys.Rev.C69:055204,2004 (Iachello et al.)
   * 
   *
   * input arguments: 
   *    q2: photon 4-momentum square)
   *  
   * output arguments:
   *    Re_GE, Im_GE: real and imaginary parts of electric Sachs FF 
   *    Re_GM, Im_GM: real and imaginary parts of magnetic Sachs FF
   *    Re_F1, Im_F1: real and imaginary parts of Pauli-Dirac F1 FF
   *    Re_F2, Im_F2: real and imaginary parts of Pauli-Dirac F2 FF
   *
   */

  double PI=acos(-1.0);

  double M=mp;                //proton mass
  double tau=q2/(4.0*M*M);    //usual "tau"
 
  double momega_local=0.784;        //omega mass (GeV) [Iachello 2004]
  double mphi_local=1.019;          //phi mass   (GeV) [Iachello 2004]
  double mrho_local=0.765;          //rho mass   (GeV) [Iachello 2004]


  double theta=53.0;          //model parameters [Iachello 2004] (see paper)
  double gamma=0.25;          //          
  double beta_omega=1.102;    // 
  double beta_phi=0.112;      //
  double beta_rho=0.672;      //
  double alpha_phi=-0.052;    //
  double Gamma_rho=0.112;     //

  double a,b,c;                        //temporal variables for calculations 
  double Re_g, Im_g;                   //intrinsic form factor g(q2)
  double P_omega, P_phi, P_rho;        //propagators terms for omega, phi, rho
  double alpha, beta;                  //alpha(q2), beta(q2) functions for modified rho propagator
  double Re_P_mod_rho, Im_P_mod_rho;   //modified rho propagator
  double Re_F1_S, Im_F1_S;             //isoscalar part of F1
  double Re_F1_V, Im_F1_V;             //isovector part of F1
  double Re_F2_S, Im_F2_S;             //isoscalar part of F2
  double Re_F2_V, Im_F2_V;             //isovector part of F2



  //check argument range
  if(q2<(4.0*mpi*mpi)){
    printf("ERROR: function mz_FF_vmd called with argument q2 out of range \n");
    printf("    => q2 should be larger than the threshold (4.0*mpi^2) = %5.3f GeV\n", (4.0*mpi*mpi));
    printf("\n\n");
    exit(1);
  }


  //intrinsic form factor g(q2)
  a=1.0;
  b=1.0-2.0*gamma*q2*cos(theta*PI/180.0)+gamma*gamma*q2*q2*cos(2.0*theta*PI/180.0);
  c=gamma*gamma*q2*q2*sin(2.0*theta*PI/180.0)-2.0*gamma*q2*sin(theta*PI/180.0);

  Re_g=(a*b)/(b*b + c*c);
  Im_g=(-1.0*a*c)/(b*b + c*c);


  //propagators for omega, phi and rho (the last one actually not used)
  P_omega=(momega_local*momega_local)/(momega_local*momega_local-q2);
  P_phi=(mphi_local*mphi_local)/(mphi_local*mphi_local-q2);
  P_rho=(mrho_local*mrho_local)/(mrho_local*mrho_local-q2);

  //modified rho-propagator
  alpha=(2.0/PI)*pow((q2-4.0*mpi*mpi)/(q2),0.5)
    *log( ( pow(q2-4.0*mpi*mpi,0.5) +pow(q2,0.5) )/(2.0*mpi) );
  beta=pow(( pow( (q2)/(4.0*mpi*mpi)-1.0,3.0) )/( (q2)/(4.0*mpi*mpi) ),0.5);

  a=mrho_local*mrho_local +8.0*Gamma_rho*mpi/PI;
  b=mrho_local*mrho_local -q2 +(4.0*mpi*mpi -q2)*Gamma_rho*alpha/mpi;
  c=Gamma_rho*4.0*mpi*beta;

  Re_P_mod_rho=(a*b)/(b*b + c*c);
  Im_P_mod_rho=(-1.0*a*c)/(b*b + c*c);


  //F1_S FF
  Re_F1_S=0.5*Re_g*( (1.0-beta_omega-beta_phi) +beta_omega*P_omega +beta_phi*P_phi);
  Im_F1_S=0.5*Im_g*( (1.0-beta_omega-beta_phi) +beta_omega*P_omega +beta_phi*P_phi);

  //F1_V FF
  Re_F1_V=0.5*( Re_g*( (1.0-beta_rho) + beta_rho*Re_P_mod_rho )  
               +(-1.0)*Im_g*beta_rho*Im_P_mod_rho );
  Im_F1_V=0.5*( Re_g*beta_rho*Im_P_mod_rho 
	       +Im_g*( (1.0-beta_rho) +beta_rho*Re_P_mod_rho ) );


  //F2_S FF
  Re_F2_S=0.5*Re_g*( (-0.120-alpha_phi)*P_omega +alpha_phi*P_phi );
  Im_F2_S=0.5*Im_g*( (-0.120-alpha_phi)*P_omega +alpha_phi*P_phi );

  //F2_V FF
  Re_F2_V=0.5*3.706*( Re_g*Re_P_mod_rho -Im_g*Im_P_mod_rho  );
  Im_F2_V=0.5*3.706*( Re_g*Im_P_mod_rho +Im_g*Re_P_mod_rho  );



  //build F1 and F2 out of isoscalar and isovector parts
  Re_F1=Re_F1_S + Re_F1_V;
  Im_F1=Im_F1_S + Im_F1_V;

  Re_F2=Re_F2_S + Re_F2_V;
  Im_F2=Im_F2_S + Im_F2_V;


  //build G_E and G_M out of F1 and F2
  Re_GM=Re_F1 + Re_F2;
  Im_GM=Im_F1 + Im_F2;

  Re_GE=Re_F1 + tau*Re_F2;
  Im_GE=Im_F1 + tau*Im_F2;


}






void mz_FF_vmd_mod(double q2, 
               double& Re_GE, double& Im_GE,
               double& Re_GM, double& Im_GM,
               double& Re_F1, double& Im_F1,
               double& Re_F2, double& Im_F2){
  /**
   * function mz_FF_pQCD 
   *
   * calculates, for a given q2, the values of both Sachs and Pauli-Dirac 
   * proton FF in the time-like region, 
   * according to the "vector meson dominance" model in
   * Phys.Rev.C69:055204,2004 (Iachello et al.)
   * modified with non-zero widths for the omega and phi-resonances
   * 
   *
   * input arguments: 
   *    q2: photon 4-momentum square)
   *  
   * output arguments:
   *    Re_GE, Im_GE: real and imaginary parts of electric Sachs FF 
   *    Re_GM, Im_GM: real and imaginary parts of magnetic Sachs FF
   *    Re_F1, Im_F1: real and imaginary parts of Pauli-Dirac F1 FF
   *    Re_F2, Im_F2: real and imaginary parts of Pauli-Dirac F2 FF
   *
   */

  double PI=acos(-1.0);

  double M=mp;                //proton mass
  double tau=q2/(4.0*M*M);    //usual "tau"
 
  double momega_local=0.784;        //omega mass (GeV) [Iachello 2004]
  double mphi_local=1.019;          //phi mass   (GeV) [Iachello 2004]
  double mrho_local=0.765;          //rho mass   (GeV) [Iachello 2004]


  double theta=53.0;          //model parameters [Iachello 2004] (see paper)
  double gamma=0.25;          //          
  double beta_omega=1.102;    // 
  double beta_phi=0.112;      //
  double beta_rho=0.672;      //
  double alpha_phi=-0.052;    //
  double Gamma_rho=0.112;     //

  double a,b,c;                        //temporal variables for calculations 
  double Re_g, Im_g;                   //intrinsic form factor g(q2)
  //double P_omega, P_phi, P_rho;        //propagators terms for omega, phi, rho
  double Re_P_omega, Im_P_omega;       //modified omega propagator with omega width       
  double Re_P_phi, Im_P_phi;           //modified phi propagator with phi width
  double Re_A, Im_A;                   //temporal variable for calculation in mod. omega/phi prop.
  double alpha, beta;                  //alpha(q2), beta(q2) functions for modified rho propagator
  double Re_P_mod_rho, Im_P_mod_rho;   //modified rho propagator
  double Re_F1_S, Im_F1_S;             //isoscalar part of F1
  double Re_F1_V, Im_F1_V;             //isovector part of F1
  double Re_F2_S, Im_F2_S;             //isoscalar part of F2
  double Re_F2_V, Im_F2_V;             //isovector part of F2



  //check argument range
  if(q2<(4.0*mpi*mpi)){
    printf("ERROR: function mz_FF_vmd called with argument q2 out of range \n");
    printf("    => q2 should be larger than the threshold (4.0*mpi^2) = %5.3f GeV\n", (4.0*mpi*mpi));
    printf("\n\n");
    exit(1);
  }


  //intrinsic form factor g(q2)
  a=1.0;
  b=1.0-2.0*gamma*q2*cos(theta*PI/180.0)+gamma*gamma*q2*q2*cos(2.0*theta*PI/180.0);
  c=gamma*gamma*q2*q2*sin(2.0*theta*PI/180.0)-2.0*gamma*q2*sin(theta*PI/180.0);

  Re_g=(a*b)/(b*b + c*c);
  Im_g=(-1.0*a*c)/(b*b + c*c);


  //propagators for omega, phi and rho (the last one actually not used)
  //P_omega=(momega_local*momega_local)/(momega_local*momega_local-q2);
  //P_phi=(mphi_local*mphi_local)/(mphi_local*mphi_local-q2);
  //P_rho=(mrho_local*mrho_local)/(mrho_local*mrho_local-q2);

  //modified rho-propagator
  alpha=(2.0/PI)*pow((q2-4.0*mpi*mpi)/(q2),0.5)
    *log( ( pow(q2-4.0*mpi*mpi,0.5) +pow(q2,0.5) )/(2.0*mpi) );
  beta=pow(( pow( (q2)/(4.0*mpi*mpi)-1.0,3.0) )/( (q2)/(4.0*mpi*mpi) ),0.5);

  a=mrho_local*mrho_local +8.0*Gamma_rho*mpi/PI;
  b=mrho_local*mrho_local -q2 +(4.0*mpi*mpi -q2)*Gamma_rho*alpha/mpi;
  c=Gamma_rho*4.0*mpi*beta;

  Re_P_mod_rho=(a*b)/(b*b + c*c);
  Im_P_mod_rho=(-1.0*a*c)/(b*b + c*c);


  //modified omega and phi-progagators with non-zero widths
  a=momega_local*momega_local;
  b=momega_local*momega_local-q2;
  c=momega_local*Gamma_omega;

  Re_P_omega=(a*b)/(b*b+c*c);
  Im_P_omega=(-1.0*a*c)/(b*b+c*c);

  a=mphi_local*mphi_local;
  b=mphi_local*mphi_local-q2;
  c=mphi_local*Gamma_phi;

  Re_P_phi=(a*b)/(b*b+c*c);
  Im_P_phi=(-1.0*a*c)/(b*b+c*c);


  //F1_S FF
  //Re_F1_S=0.5*Re_g*( (1.0-beta_omega-beta_phi) +beta_omega*P_omega +beta_phi*P_phi);
  //Im_F1_S=0.5*Im_g*( (1.0-beta_omega-beta_phi) +beta_omega*P_omega +beta_phi*P_phi);
  Re_A=(1.0-beta_omega-beta_phi) +beta_omega*Re_P_omega +beta_phi*Re_P_phi;
  Im_A=                           beta_omega*Im_P_omega +beta_phi*Im_P_phi;

  Re_F1_S=0.5*(Re_g*Re_A -Im_g*Im_A);
  Im_F1_S=0.5*(Re_g*Im_A +Im_g*Re_A);


  //F1_V FF
  Re_F1_V=0.5*( Re_g*( (1.0-beta_rho) + beta_rho*Re_P_mod_rho )  
               +(-1.0)*Im_g*beta_rho*Im_P_mod_rho );
  Im_F1_V=0.5*( Re_g*beta_rho*Im_P_mod_rho 
	       +Im_g*( (1.0-beta_rho) +beta_rho*Re_P_mod_rho ) );


  //F2_S FF
  //Re_F2_S=0.5*Re_g*( (-0.120-alpha_phi)*P_omega +alpha_phi*P_phi );
  //Im_F2_S=0.5*Im_g*( (-0.120-alpha_phi)*P_omega +alpha_phi*P_phi );
  Re_A=(-0.120-alpha_phi)*Re_P_omega +alpha_phi*Re_P_phi;
  Im_A=(-0.120-alpha_phi)*Im_P_omega +alpha_phi*Im_P_phi;

  Re_F2_S=0.5*(Re_g*Re_A -Im_g*Im_A);
  Im_F2_S=0.5*(Re_g*Im_A +Im_g*Re_A);


  //F2_V FF
  Re_F2_V=0.5*3.706*( Re_g*Re_P_mod_rho -Im_g*Im_P_mod_rho  );
  Im_F2_V=0.5*3.706*( Re_g*Im_P_mod_rho +Im_g*Re_P_mod_rho  );



  //build F1 and F2 out of isoscalar and isovector parts
  Re_F1=Re_F1_S + Re_F1_V;
  Im_F1=Im_F1_S + Im_F1_V;

  Re_F2=Re_F2_S + Re_F2_V;
  Im_F2=Im_F2_S + Im_F2_V;


  //build G_E and G_M out of F1 and F2
  Re_GM=Re_F1 + Re_F2;
  Im_GM=Im_F1 + Im_F2;

  Re_GE=Re_F1 + tau*Re_F2;
  Im_GE=Im_F1 + tau*Im_F2;


}




void mz_FF_Tomasi(double q2, 
                double& Re_GE, double& Im_GE,
                double& Re_GM, double& Im_GM,
                double& Re_F1, double& Im_F1,
                double& Re_F2, double& Im_F2){
  /**
   * function mz_FF_Tomasi 
   *
   * calculates, for a given q2, the values of both Sachs and Pauli-Dirac 
   * proton FF, according to the dipole-modified model in
   * Eur. Phys. J A 44, 373-384 (2010) (Sudol et al.) eq. 7 
   *
   * input arguments: 
   *    q2: photon 4-momentum square)
   *  
   * output arguments:
   *    Re_GE, Im_GE: real and imaginary parts of electric Sachs FF 
   *    Re_GM, Im_GM: real and imaginary parts of magnetic Sachs FF
   *    Re_F1, Im_F1: real and imaginary parts of Pauli-Dirac F1 FF
   *    Re_F2, Im_F2: real and imaginary parts of Pauli-Dirac F2 FF
   *
   */

  double PI=acos(-1.0);

  double A=22.5;
  double B=0.71;
  double C=3.6;
  double G;

  G=A/(1.0+q2/B)/(1.0+q2/B)/(1.0+q2/C);

  Re_GE=G;
  Im_GE=0.0;

  Re_GM=G;   
  Im_GM=0.0;

  Re_F1=G;
  Im_F1=0.0;

  Re_F2=0.0;
  Im_F2=0.0;


}


