//c/c++
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>

//ranlux
#include "ranlxs.h"
#include "ranlxd.h"

void ranlxs(float r[], int n);
void ranlxd(double r[], int n);

void rlxs_init(int level, int seed);
void rlxd_init(int level, int seed);


//mz general
#include "mzparameters.h"
#include "mzfunctions.h"

//mz for sigma_leplep_sigma_pipi_vandewi
#include "mzfunctions_sigma_leplep_sigma_pipi_vandewi.h"

//mz for pp_to_leplep_vandewi
#include "mzfunctions_pp_to_leplep_vandewi.h"

//mz for pp_to_pipi_vandewi
//#include "pipisigmas.h"
#include "mzfunctions_pp_to_pipi_vandewi.h"


using namespace std;


void mz_sigma_leplep_sigma_pipi_vandewi_init(int particle_flag, double P, 
                                             double cos_theta_min, 
                                             double cos_theta_max,
                                             double FF_flag, int seed)
{
  /**
   * function mz_pp_to_leplep_vandewi_init
   * 
   * INPUT:
   *
   *  seed            -> RANLUX seed
   *  particle_flag   -> 0=electron, 1=muon, 2=tau
   *  P               -> anti-proton momentum (lab frame) [GeV]
   *  FF_flag         -> flag for Form Factor model: 0=pQCD, 1=Iachello
   *  cos_theta_min   -> cos(theta*) minimum; theta*=angle(lep-,pbar) CMS frame
   *  cos_theta_max   -> cos_theta_max: cos(theta*) maximum
   *
   * OUTPUT:
   *
   *  without output
   *
   *
   * DESCRIPTION:
   *
   *  function  mz_pp_to_leplep_vandewi_init checks the values of 
   *  all arguments, killing the job if at least one value lies 
   *  outside its allowed range.
   *  In case of tau production, it determines compatibility 
   *  with the input antiproton momentum, killing the job if 
   *  the proton-antiproton center-of-mass energy does not reach the threshold
   *  for tau production. Kinematic information is printed out to the log file. 
   *  RANLUX is initialised.
   *
   */

#include "pipisigmas.h"

  double P_PANDA_THRESHOLD=1.5;        //pbar momentum threshold in PANDA
  double P_PANDA_UP=15.0;              //pbar momentum upper limit in PANDA

  double M=mp;             //proton mass

  double E;                //lab frame: pbar energy
  double S;                //center-of-mass energy squared of pbarpsystem
  double sqrt_S;           //center-of-mass energy of pbarpsystem
  double S_min;            //threshold for tau+tau- production: center-of-mass energy squared of pbarpsystem 
  double E_min;            //threshold for tau+tau- production: (lab frame) pbar energy 
  double P_min;            //threshold for tau+tau- production: (lab frame) pbar momentum


  double p_a=pz[0];        //momentum range for pion cross section
  double p_b=pz[19];
  double p_c=pz[20];
  double p_d=pz[38];

  int p_flag;              //pion cross section: momentum extrapolation flag
  int i_min;               //pion cross section: lower site in p-lattice for input momentum P


  //checking parameter range...

  //lepton flag
  if( !( ( particle_flag==0)||( particle_flag==1)||( particle_flag==2) ) ){
    printf("ERROR: particle flag out of range: \n");
    printf("    => particle flag should be 0 (=electron), 1 (=muon) or 2 (=tau)\n");
    printf("    => job killed!!! \n");
    printf("\n\n");
    exit(1);
  }


  //antiproton momentum: general
  if(P<0.0){
    printf("ERROR: P (anti-proton momentum) out of range: \n");
    printf("    => P should be a positive real number \n");
    printf("    => job killed!!! \n");
    printf("\n\n");
    exit(1);
  }


  //antiproton momentum: PANDA threshold
  if(P<P_PANDA_THRESHOLD){
    printf("WARNING: P (anti-proton momentum) smaller than PANDA threshold %4.2f GeV \n", 
           P_PANDA_THRESHOLD);
    printf("\n\n");
  }


  //antiproton momentum: PANDA upper limit
  if(P>P_PANDA_UP){
    printf("WARNING: P (anti-proton momentum) larger than PANDA upper limit %4.2f GeV \n", 
           P_PANDA_UP);
    printf("\n\n");
  }


  /**
   * anti proton momentum: initial state kinematics and tau production...
   */

       //antiproton energy
       E=pow((P*P+M*M),0.5);

       //center-of-mass energy squared and center-of-mass energy of pbarpsystem
       S=2.0*M*(M+E);
       sqrt_S=pow(S,0.5);

       //kinematic threshold for tau production
       S_min=4.0*mtau*mtau;
       E_min=S_min/(2.0*M)-M;
       P_min=pow((E_min*E_min-M*M),0.5);

       //check if tau production is possible
       if( (particle_flag==2)&&(S<=S_min) ){
        printf("ERROR: center of mass energy below threshold for tau+tau- production \n");
        printf("    => tau+tau- production requires: \n");
        printf("           center-of-mass energy squared   s > 4.0*m_{tau}^2 = %4.2f GeV^2, i.e. \n", S_min);
        printf("           antiproton energy               E > %4.2f GeV, i.e. \n", E_min);
        printf("           antiproton momentum             P > %4.2f GeV \n", P_min);
        printf("    => job killed!!! \n");
        exit(1);
        printf("\n\n");
       }


  //antiproton momentum: pion cross section
  if( !( (P>=p_a)&&(P<=p_d) ) ){
    printf("ERROR: P (anti-proton momentum) out of range: no pion cross section available \n");
    printf("    => P should be contained in the interval [%4.2f,%5.2f] GeV\n",p_a,p_d);
    printf("    => job killed!!! \n");
    printf("\n\n");
    exit(1);
  }


  //antiproton momentum and cos(theta*) correlations: pion cross section
  if( ((P>=p_a)&&(P<p_c))&&( (cos_theta_min<-0.94)||(cos_theta_max>0.94)) ){
    printf("WARNING: in the momentum range [%4.2f,%5.2f) GeV,\n",p_a,p_c);
    printf("         it is recommended to keep cos(theta*) range within the interval [-0.94,0.94]\n");
    printf("         to avoid potentially large extrapolations in the pion cross section\n");
    printf("\n\n");
  }



  //pion production: determining kinematic range and cross section parametrization...

  printf("PION PRODUCTION: Kinematic regime and cross section parametrization\n");
  printf("-------------------------------------------------------------------\n");
  printf("\n");

  if( (P>=p_a)&&(P<=p_b)  ){
    printf("%5.2f < P < %5.2f GeV => low energy regime\n", p_a, p_b);
    printf("cross section parametrization: Legendre polynomial series\n");
    printf("\n\n");
  }

  if( (P>p_b)&&(P<p_c)  ){
    printf("%5.2f < P < %5.2f GeV => GAP energy regime\n", p_b, p_c);
    printf("cross section parametrization: extrapolation Legendre polynomial series / Regge description\n");
    printf("\n\n");
  }

  if( (P>=p_c)&&(P<=pz[24])  ){
    printf("%5.2f < P < %5.2f GeV => transition energy regime\n", p_c, pz[24]);
    printf("cross section parametrization: Regge description\n");
    printf("\n\n");
  }


  if( (P>pz[24])&&(P<=p_d)  ){
    printf("%5.2f < P < %5.2f GeV => high energy regime\n", pz[24],p_d);
    printf("cross section parametrization: Regge description\n");
    printf("\n\n");
  }


  //pion  production: determining if some cross section extrapolation is needed...

  mz_pp_to_pipi_get_p_lattice_site(P, i_min, p_flag);


  if(p_flag==0){
    printf("P at lattice site: no cross section interpolation needed \n");
    printf("\n\n\n\n");
  }

  if(p_flag==1){
    printf("P NOT at lattice site: interpolating cross section between momenta %5.2f and %5.2f GeV\n", pz[i_min], pz[i_min+1]);
    printf("\n\n\n\n");
  }



       
  //cos(theta*)
  if( !( (cos_theta_min>=-1.0)&&(cos_theta_max>cos_theta_min)&&(cos_theta_max<=1.0) ) ){
    printf("ERROR: cos(theta*) range NOT allowed \n");
    printf("    => cos(theta*) range [cos(theta*) min, cos(theta*) max]\n");
    printf("       should be contained within [-1.0,1.0]\n");
    printf("    => job killed!!! \n");
    printf("\n\n");
    exit(1);
  }


  //FF models
  if( !( (FF_flag==0)||(FF_flag==1)||(FF_flag==2) ) ){
    printf("ERROR: FF flag out of range: \n");
    printf("    => FF flag should be 0 (=pQCD), 1 (=vmd) or 2 (=dipole-modified)\n");
    printf("    => job killed!!! \n");
    printf("\n\n");
    exit(1);
  }


  //RANLUX
  if( (seed<1)||(seed>2147483647) ){   
   printf("ERROR: RANLUX seed out of range: \n");
   printf("    => RANLUX seed should be contained within [1,2^31-1]\n");
   printf("    => job killed!!! \n");
   printf("\n\n");
   exit(1);
  }




  //print out kinematics
  printf("Initial state kinematics\n");
  printf("------------------------\n");
  printf("\n");                                
  printf("=> CM energy square:        S = %5.2f GeV^2 \n", S);
  printf("   CM energy:         sqrt(S) = %5.2f GeV \n", sqrt_S);
  printf("\n\n\n\n");




  //initialize RANLUX
  rlxd_init(1,seed);


}

